c     CHEMSEP Copyright (c) H.A.Kooijman R.Taylor 1988-2021 All rights reserved

      program TestCOPP
c
c     Test program for ChemSep CSCOPP.dll
c
      character        SepFile*250, Msg*250, Phases*5
      integer          nc,nap,iErr,io
      integer          ncmax
      parameter       (ncmax=40)
      integer          i, j, k, iLib, iOpt(4)
      character        Cname*50, Desc*50
      character        CAS*12, SStruct*50, Smiles*171
      double precision Mw,x(ncmax),y(ncmax),z(ncmax)
      double precision T,p,Cvalue
      integer          iPhase, iFlsh
      integer          StrLen, nd_copp
      double precision Q,Ffeed,Vapour,Liquid,Sigma
      double precision MolwtL, RhoL, CL, VisL, ConL, CpL
      double precision MolwtV, RhoV, CV, VisV, ConV, CpV
      double precision Dvap(ncmax,ncmax)
      double precision Dliq(ncmax,ncmax)
      double precision Kval(ncmax)
      double precision dKdx(ncmax,ncmax),dKdy(ncmax,ncmax)
      double precision dKdT(ncmax),dKdP(ncmax)
      double precision LnPhiV(ncmax),LnPhiL(ncmax),dLnPhi(ncmax,ncmax)
      double precision dPhiT(ncmax),dPhiP(ncmax)
      double precision dHdx(ncmax),dHdT,dHdP
      double precision S,dSdx(ncmax),dSdT,dSdP
      double precision HV, HL, Hfeed
      double precision Toler
      integer          nex, iDmns(7)
      integer          NumberExtraProps, idExtraProperty, IsActCof

c      include 'koptions.inc'
      integer         Kmodel
      common /Kopts/  Kmodel(6)
c      include 'hoptions.inc'
      integer           Hmodel
      common /Hopts/    Hmodel
      double precision  Hscale
      common /Hopts2/   Hscale
c      include 'hreference.inc'
      integer           iHrefS
      integer           iHform
      integer           iHpath
      double precision  HrefT
      common /HEATSI/   iHrefS, iHform
      common /HEATS3/   HrefT
      common /HeatsP/   iHpath

      write(*,*) 'TESTCOPP started'
c     - set output to 6 to view results, set to 0 to verify the fortran doesn't write failures to stdout -
      io = 6
      SepFile = 'd:\ChemSep\validate\flowsheets\' //
     +   'P_Cativa\Helium-Compressor_1.sep'
      iErr = 1
      Toler=1.0d-9
      iOpt(1)=1
      iOpt(2)=1
      iOpt(3)=1
      iOpt(4)=1

c     ------------------------------------------------------------------
c     integer function nd_copp
c     Returns the dimension of arrays/matrices in the cscopp.dll
c     ------------------------------------------------------------------
      i=nd_copp()
      if (ncmax .ne. i) then
        write(*,*) 'Wrong ncmax definition in COPP ',i,' vs ',ncmax
        stop
      end if

c     ------------------------------------------------------------------
c     subroutine iCScopp(SepFil,Msg,nap,nc,Toler,iErr)
c     Opens the Sep-File and reads input, assigns all model selections and model parameters
c     Input:
c         SepFile
c     Output:
c         Msg   = License+Build / Error
c         nap   = Number of apparent species at Cape-Open level
c         nc    = Number of true components
c         iErr  = error flag (error if <>0)
c     character SepFil*(*), Msg*(*)
c     integer   nc,iErr
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call iCScopp(SepFile,Msg,nap,nc,iErr)
      if (iErr .ne. 0) then
        j=StrLen(Msg)
        write(*,*) '*** Failure in COPP: error = ',Msg(1:j)
        stop
      else if (io .ne. 0) then
        j=StrLen(SepFile)
        write(io,*) 'Initialized ChemSep COPP from ',SepFile(1:j)
        j=StrLen(Msg)
        write(io,*) Msg(1:j)
        write(io,*) 'Number of apparent components = ',nap
        write(io,*) 'Number of true components = ',nc
      end if
c      call DebugOn('btx.dbg',iErr)

c     ------------------------------------------------------------------
c     subroutine GetPhases(Phases,iErr)
c     Determine the phases in the property package
c     Output:
c         Phases = String(5) with VLE, LLE, VLLE, VLSE, VLLSE
c         iErr  = error flag, 0 on success
c     character Phases*(*)
c     integer iErr
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call GetPhases(Phases,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in GetPhases: error = ',iErr
        stop
      else if (io .ne. 0) then
        write(io,*) 'Phase equilibria ChemSep COPP = ',Phases
      end if

c     ------------------------------------------------------------------
c     subroutine Get_Comp(Cname,CAS,Smiles,SStruc,Mw,iLib,i,ncs,nd,iErr)
c     Get component's i information
c     Input:
c         i = Component to request info on
c         nc = number of components
c         nd = dimension of component arrays/matrices
c     Output:
c         Cname = Component name
c         CAS   = CAS number (string with -)
c         Smiles= Smiles string
c         SStruc= Chemical structure string
c         iLib  = library number
c         Mw    = Mole weight (kg/kmol)
c         iErr  = error flag, 0 on success
c     integer   iLib,i,nc,nd,iErr
c     character Cname*(*)
c     character CAS*(*)
c     double precision Mw
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      do i=1,nap
        call Get_Comp(Cname,CAS,Smiles,SStruct,Mw,iLib,i,nc,ncmax,iErr)
        if (iErr .ne. 0) then
          write(*,*) '*** Failure in Get_Comp: error = ',iErr
        else if (io .ne. 0) then
          j=StrLen(Cname)
          write(io,*) 'Component ',i,'=',Cname(1:j)
          j=StrLen(CAS)
          write(io,*) 'iLib=',iLib,' CAS=',CAS(1:j),' MW= ',Mw
          j=StrLen(SStruct)
          k=StrLen(Smiles)
          write(io,*) 'Structure=',SStruct(1:j),' SMILES=',Smiles(1:k)
        end if
      end do


c     ------------------------------------------------------------------
c     subroutine Get_Const(Cvalue,iConst,i,iErr)
c     Request pure component data for component i
c     Input:
c          i      = component to request information on
c          iConst = the property to report, where order is as in our PCD file: Tc=1, Pc=2, ... 42
c           1: Tc (K)                              22: H combustion (J/kmol)
c           2: Pc (Pa)                             23: V* (m3/kmol)
c           3: Vc (m3/mol)                         24: Lennard-Jones sigma (m)
c           4: Zc                                  25: Lennard-Jones epsilon (K)
c           5: T normal boiling (K)                26: Rackett parameter
c           6: T melt (K)                          27: Fuller diffusion volume (cm3)
c           7: T tripple (K)                       29: Parachor
c           8: T tripple (Pa)                      30: Specific gravity
c           9: Molecular weight (kg/kmol)          31: Charge
c          10: Vboil (m3/kmol)                     32: SRK acentric factor
c          11: Acentric factor                     33: Wilson volume (m3/kmol)
c          12: Radius of gyration (m)              34: UNIQUAC R
c          13: Solubility parameter sqrt(J/m3)     35: UNIQUAC Q
c          14: Dipole moment (Coulomb . m)         36: UNIQUAC Q'
c          15: van der Waals Volume (m3/kmol)      37: API SRK parameter s1
c          16: van der Waals Area (m2/kmol)        38: API SRK parameter s2
c          17: H formation (J/kmol)                40: Chao-Seader acentric factor
c          18: G formation (J/kmol)                41: Chao-Seader solubility parameter sqrt(J/m3)
c          19: Absolute Entropy (J/kmol/K)         42: Chao-Seader volume (m3/kmol)
c          20: H fusion at Tmelt (J/kmol)          28: Mathias parameter C2
c          21: Mathias parameter C1                39: Mathias parameter C3
c      Output:
c          Cvalue = property value requested
c         iErr  = error flag
c                 0 on success,
c                 1 if un-initialized,
c                 2 if i>nc,
c                 3 if iConst<1 or iConst>50,
c                 4 if value is not available
c      integer   iConst,i,iErr
c      double precision Cvalue
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      do i=1,nap
        call Get_Const(Cvalue,2,i,iErr)
        if (iErr .ne. 0) then
          write(*,*) '*** Failure in Get_Const: error = ',iErr
        else if (io .ne. 0) then
          write(io,*) 'Pc(',i,')= ',Cvalue
        end if
      end do
c
      i = 1
      call Get_Const(Cvalue,1,i,iErr)
      if ((iErr .eq. 0) .and. (io .ne. 0)) then
        write(io,*) 'Tc(',i,')= ',Cvalue
      end if
      call Get_Const(Cvalue,2,i,iErr)
      if ((iErr .eq. 0) .and. (io .ne. 0)) then
        write(io,*) 'Pc(',i,')= ',Cvalue
      end if
      call Get_Const(Cvalue,3,i,iErr)
      if ((iErr .eq. 0) .and. (io .ne. 0)) then
        write(io,*) 'Vc(',i,')= ',Cvalue
      end if
      call Get_Const(Cvalue,4,i,iErr)
      if ((iErr .eq. 0) .and. (io .ne. 0)) then
        write(io,*) 'Zc(',i,')= ',Cvalue
      end if

c     ------------------------------------------------------------------
c      subroutine C_Flash(IFTYPE,IMETH,Q,Ffeed,
c     +                  iPhase,Vapour,Liquid,
c     +                  NEEDCOMP0PHASE,
c     +                  T,P,Z,x,y,ncs,nd,iErr)
C     Input:
C          IFTYPE     Identifies flash type
C                     1 - QP
C                     2 - TP
C                     3 - PV
C                     4 - QT
C                     5 - QV
C                     6 - TV
C                     7 - pS
C                     8 - PH
C          IMETH      Designates numerical method used
C                     IMETH < 0 uses Rachford-Rice method
C                     IMETH >=0 uses Newton's method
C                     mole fractions of component i in feed
C          NC         Number of components
C          Z          Array of length NC containing
C                     mole fractions of component i in feed
C          T          Temperature (Kelvin)
C          P          Pressure (Pascals)
C          FFeed      Feed flow rate or amount
C          NEEDCOMP0PHASE Compositions are required for vanishing
C                     phase in bub- or dewpoint calculation
C     Output:
C          iPhase     Phase flag (0 = two phase, 1 = light only, 2 = dense only)
C          Vapour     Vapour flow rate or amount
C          Liquid     Liquid flow rate or amount
C          X          Array of length NC containing
C                     mole fractions of component i in Liquid
C          Y          Array of length NC containing
C                     mole fractions of component i in vapor
C          T          Temperature (Kelvin)
C          P          Pressure (Pascals)
c         iErr  = error flag
c                 0 on success,
c                 -1 not initialized
c                 >0 reached maximum number of iterations w/o converging
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      z(1) = 1.d0
      z(2) = 0.0d0
      z(3) = 0.0d0
      x(1) = z(1)
      x(2) = z(2)
      x(3) = z(3)
      y(1) = z(1)
      y(2) = z(2)
      y(3) = z(3)
      Q = -5698.96d0
      p = 5.0d5
      Vapour=0.0d0
      Ffeed=1.d0
c     QP flash
      iFlsh=7
c
      call C_Flash(iFlsh,0,Q,Ffeed,
     +             iPhase,Vapour,Liquid,
     +             1,
     +             T,P,Z,x,y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in C_Flash: error = ',iErr
        stop
      else if (io .ne. 0) then
        write(io,*) 'Calculated ChemSep Flash'
        write(io,*) 'Flash type: ', iFlsh
        write(io,*) 'T =',T,' p = ',P
        write(io,*) 'F =',Ffeed,' L = ',Liquid,' V = ',Vapour
        write(io,*) 'Q =',Q
        do i=1,nap
          write(io,*) 'i =',i,' x(i)=',x(i),' y(i)=',y(i)
        end do
      end if

c     ------------------------------------------------------------------
c      subroutine PProp_L(MOLWTL, RHOL,  CL, VISL,  CONL,  CPL,
c     +                   JMOLWT, JRHOL,     JVISL, JCONL, JCPL,
c     +                   T,P,X,nc,nd,iErr)
c     Input:
c        JXXXXX - Flag to indicate what is to be computed (0 to ignore)
c                 CL only computed if JRHOL=1
c        T      - Temperature (Kelvin)
c        P      - Pressure (Pascals)
c        X      - Vector with mixture composition (mole fraction)
c        NC     - Number of components
c        ND     - Dimension of all arrays in argument list
c     Output:
c        MOLWTL - Molecular weight (kg / kmol)
c        RHOL   - Liquid density (kg / cubic meter)
c        CL     - Liquid molar density (kmol / cubic meter)
c        VISL   - Liquid viscosity (Pascal seconds)
c        CONL   - Liquid thermal conductivity
c        CPL    - Liquid heat capacity (J / kmol)
c     integer            NC, ND
c     integer            JMOLWT, JRHOL, JVISL, JCONL, JCPL
c     double precision   T, P, X(ND)
c     double precision   MOLWTL, RHOL,  CL, VISL,  CONL,  CPL, MWL
c     integer            EOSMOD
c     ------------------------------------------------------------------

      if (io .ne. 0) write(io,*)
      call PProp_L(MOLWTL, RHOL,  CL, VISL,  CONL,  CPL,
     +             1,      1,         1,     1,     1,
     +             T,P,X,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in PProp_L: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'PProp_L succeeded'
        write(io,*) 'Mw=',MolwtL,' RhoL=',RhoL,' CL=',Cl
        write(io,*) 'VisL=',VisL,' ConL=',ConL,' CpL=',CpL
      end if

c     ------------------------------------------------------------------
c      subroutine PProp_V(MOLWTV, RHOV,  CV, VISV,  CONV,  CPV,
c     +                   JMOLWT, JRHOV,     JVISV, JCONV, JCPV,
c     +                   T,P,Y,nc,nd,iErr)
c     Input:
C        JXXXXX - Flag to indicate what is to be computed
c                 CV only computed if JRHOV=1
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pascals)
C        X      - Vector with mixture composition (mole fraction)
C        NC     - Number of components
C        ND     - Dimension of X in argument list
C     Output:
C        MOLWTV - Molecular weight (kg / kmol)
C        RHOV   - Vapour density (kg / cubic meter)
C        VISV   - Vapour viscosity (Pascal seconds)
C        CONV   - Vapour thermal conductivity
C        CPV    - Vapour heat capacity (J / kmol)
C        CV     - Vapour mixture molar density (kmol / cubic meter)
C        iErr   - Flag <> 0 if vapor density has been reset
c     integer            NC, ND
c     integer            JMOLWT, JRHOV, JVISV, JCONV, JCPV
c     integer            KMODEL(6)
c     integer            iErr
c     double precision   T, P, X(ND)
c     double precision   MOLWTV, RHOV,  CV, VISV,  CONV,  CPV, MWV
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call PProp_V(MOLWTV, RHOV,  CV, VISV,  CONV,  CPV,
     +             1,      1,         1,     1,     1,
     +             T,P,Y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in PProp_V: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'PProp_V succeeded'
        write(io,*) 'MwV=',MolwtV,' RhoV=',RhoV,' CV=',CV
        write(io,*) 'VisV=',VisV,' ConV=',ConV,' CpV=',CpV
      end if

c     ------------------------------------------------------------------
c      subroutine Diff_V(Dvap, MolwtV, RhoV, VisV,
c     +                  T,P,Y,nc,nd,iErr)
C     Input:
C        MolwtV - Molecular weight (kg /kmol)
C        RhoV   - Density (kg / cubic meter)
C        VisV   - Vapor viscosity (Pa.s) only needed for high pressures
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pascals)
C        X      - Vector with mixture composition (mole fraction)
C        NC     - Number of components
C        ND     - Dimension of X in argument list
C     Output:
C        DVAP   - Vapour diffusivities (meter square / seconds)
c     integer          nc,nd,Kmodel(6)
c     double precision T, P
c     double precision X(ND), Dvap(ND,ND)
c     double precision VISV, RHOV
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call Diff_V(Dvap, MolwtV, RhoV, VisV,
     +            T,P,Y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Diff_V: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Diff_V succeeded'
        do i=1,nap
          write(io,*) 'DV=',i,(Dvap(i,j),j=1,nap)
        end do
      end if

c     ------------------------------------------------------------------
c      subroutine Diff_L(Dliq, T,P,X,nc,nd,iErr)
C     Input
C     -----
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pascals)
C        X      - Vector with mixture composition (mole fraction)
C        NC     - Number of components
C        ND     - Dimension of all arrays in argument list
C
C     Output
C     ------
C        DLIQ   - Liquid phase diffusion coefficients (square meters/s)
C
c     integer            nc, nd
c     double precision   T, P, X(nd)
c     double precision   Dliq(nd,nd)
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call Diff_L(Dliq, T,P,X,VisL,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Diff_L: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Diff_L succeeded'
        do i=1,nap
          write(io,*) 'DL=',i,(Dliq(i,j),j=1,nap)
        end do
      end if

c     ------------------------------------------------------------------
c      subroutine IntfTen(Sigma, RhoLiq, RhoV1,
c     +                   T,P,X,Y,nc,nd,iErr)
C     Input:
C        X      - Array of working length NC whose i-th element contains
C                 the mole fraction of component i in phase 'x'
C        Y      - Array of working length NC whose i-th element contains
C                 the mole fraction of component i in phase 'y'
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pa) (only if ISIGM = 2)
C        RhoLiq - MOLAR Liquid density (kmol/m^3) (only for some methods)
C        RhoV1  - MOLAR Vapor/2nd Liquid density (kmol/m^3) (only for some methods)
C        NC     - Number of components
C        NX     - Dimension of arrays in argument list
C     Output:
C        Sigma  - Mixture surface tension (N/m)
c     integer            iPhase, NC, NX
c     double precision   T, P
c     double precision   Sigma
c     double precision   RhoLiq, RhoV1
c     double precision   x(nx), y(nx)
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call IntfTen(Sigma,RhoL,RhoV,T,P,X,Y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in IntfTen: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'IntfTen succeeded'
        write(io,*) 'sigma=',Sigma
      end if

c     ------------------------------------------------------------------
c      subroutine K_Value(Kval,dKdx,dKdy,dKdT,dKdP,iOpt,
c     +                   T,P,X,Y,nc,nd,iErr)
C     Input:
C        X      - Vector with mixture composition (mole fraction)
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pascals)
C        NC     - Number of components
C        iOpt   - integer vector of length 4 containing flags to
C                 indicating which properties are to be evaluated
C                 iOpt(1) > 0: K-values
C                 iOpt(2) > 0: Composition derivatives
C                 iOpt(3) > 0: T derivatives
C                 iOpt(4) > 0: P derivatives
C     Output:
C        Kval   - Vector of working length NC containing K-values
C        dKdx   - Array of working area NC by NC containing
C                 partial derivatives of K-values with respect to
C                 X-phase composition
C        dKdy   - Array of working area NC by NC containing
C                 partial derivatives of K-values with respect to
C                 Y-phase composition
C        dKdT   - Vector of working length NC containing
C                 partial derivatives of K-values with respect to
C                 temperature
C        dKdP   - Vector of working length NC containing
C                 partial derivatives of K-values with respect to
C                 pressure
c     integer          NC, ND, Kmod(6), iOpt(4)
c     double precision T, P
c     double precision Kval(ND), dKdP(ND), dKdT(ND)
c     double precision dKdx(ND,nd), dKdy(ND,nd)
c     double precision x(nd), y(nd)
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call K_Value(Kval,dKdx,dKdy,dKdT,dKdP,iOpt,
     +             T,P,X,Y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in K_Value: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'K_Value succeeded'
        do i=1,nap
          write(io,*) 'K (',i,')=',Kval(i)
        end do
        if (iOpt(2) .ne. 0) then
          do i=1,nap
            write(io,*) 'dKdx=',(dKdx(i,j),j=1,nap)
          end do
          do i=1,nap
            write(io,*) 'dKdy=',(dKdy(i,j),j=1,nap)
          end do
        end if
        if (iOpt(3) .ne. 0) then
          do i=1,nap
            write(io,*) 'dKdT=',dKdt(i)
          end do
        end if
        if (iOpt(4) .ne. 0) then
          do i=1,nap
            write(io,*) 'dKdP=',dKdP(i)
          end do
        end if
      end if

c     ------------------------------------------------------------------
c      subroutine Fug_Cof(lnphi,dlnphi,DPHIT,DPHIP,iPhase,iOpt,
c     +                   T,P,x,nc,nd,iErr)
C     Input:
C        X      - Vector with mixture composition (mole fraction)
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pascals)
C        NC     - Number of components
C        iPhase - Phase for which properties are evaluated
C                 iPhase = 1: vapor phase
C                 iPhase = 2: liquid phase
C        iOpt   - Integer vector of length 4 containing flags to
C                 indicating which properties are to be evaluated
C                 iOpt(1) > 0: Fugacity coefficients
C                 iOpt(2) > 0: Composition derivatives
C                 iOpt(3) > 0: T derivatives
C                 iOpt(4) > 0: P derivatives
C        nd     - Dimension of all arrays in argument list
C     Output:
C        Remaining quantities calculated only as specified by iOpt
C        lnphi  - Vector of working length NC containing the
C                 natural logarithm of the fugacity coefficients
C        dlnphi - Array of working area NC by NC containing the
C                 partial derivatives of the natural logarithm of
C                 the fugacity coefficients with respect to mole fraction.
C                 These are the UNCONSTRAINED mole fraction derivatives,
C                 the derivatives in which the mole fractions are
C                 NOT required to sum to unity.
C        DPHIT  - Vector of working length NC containing the
C                 derivatives of ln phi(i) with respect to
C                 temperature.
C        DPHIP  - Vector of working length NC containing the
C                 derivatives of ln phi(i) with respect to
C                 pressure
c     integer            nc, nd, Iphase
c     integer            iOpt(4)
c     integer            Kmod(6)
c     double precision   lnphi(nd)
c     double precision   dlnphi(nd,nd)
c     double precision   DPHIT(nd)
c     double precision   DPHIP(nd)
c     double precision   T, P, x(nd)
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call Fug_Cof(LnPhiV,dLnPhi,dPhiT,dPhiP,1,iOpt,
     +             T,P,Y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Fug_Cof: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Fug_Cof succeeded'
        do i=1,nap
          write(io,*) 'ln(phi)V (',i,')=',LnPhiV(i)
        end do
        if (iOpt(2) .ne. 0) then
          do i=1,nap
            write(io,*) 'dLn(phi)dx=',(dLnPhi(i,j),j=1,nap)
          end do
        end if
        if (iOpt(3) .ne. 0) then
          do i=1,nap
            write(io,*) 'dLn(phi)dT=',dPhiT(i)
          end do
        end if
        if (iOpt(4) .ne. 0) then
          do i=1,nap
            write(io,*) 'dLn(Phi)dP=',dPhiP(i)
          end do
        end if
      end if
      call Fug_Cof(LnPhiL,dLnPhi,dPhiT,dPhiP,2,iOpt,
     +             T,P,X,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Fug_Cof: error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Fug_Cof succeeded'
        do i=1,nap
          write(io,*) 'ln(phi)L (',i,')=',LnPhiL(i)
        end do
        if (iOpt(2) .ne. 0) then
          do i=1,nap
            write(io,*) 'dLn(phi)dx=',(dLnPhi(i,j),j=1,nap)
          end do
        end if
        if (iOpt(3) .ne. 0) then
          do i=1,nap
            write(io,*) 'dLn(phi)dT=',dPhiT(i)
          end do
        end if
        if (iOpt(4) .ne. 0) then
          do i=1,nap
            write(io,*) 'dLn(Phi)dP=',dPhiP(i)
          end do
        end if
      end if
c
      if (ierr .eq. 0 .and. io .ne. 0) then
        write(io,*) 'Check on V & L Fug_Cof with K values'
        write(io,*) 'i    ln(phi)L    ln(phi)V    phiL/phiV     K'
        do i=1,nap
          Cvalue = LnPhiL(i) - LnPhiV(i)
          Cvalue = dexp( Cvalue )
          write(io,*) i,LnPhiL(i),LnPhiV(i),Cvalue,Kval(i)
        end do
      end if

c     ------------------------------------------------------------------
c      subroutine Act_Cof(lnGam,dGamdx,dGamT,dGamP,iPhase,iOpt,
c     +                   T,P,x,nc,nd,iErr)
C     Input:
C        X      - Vector with mixture composition (mole fraction)
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pascals)
C        NC     - Number of components
C        iPhase - Phase for which properties are evaluated
C                 iPhase = 1: vapor phase
C                 iPhase = 2: liquid phase
C        iOpt   - Integer vector of length 4 containing flags to
C                 indicating which properties are to be evaluated
C                 iOpt(1) > 0: Fugacity coefficients
C                 iOpt(2) > 0: Composition derivatives
C                 iOpt(3) > 0: T derivatives
C                 iOpt(4) > 0: P derivatives
C        nd     - Dimension of all arrays in argument list
C     Output:
C        Remaining quantities calculated only as specified by iOpt
C        lnGam  - Vector of working length NC containing the
C                 natural logarithm of the fugacity coefficients
C        dGamdx - Array of working area NC by NC containing the
C                 partial derivatives of the natural logarithm of
C                 the fugacity coefficients with respect to mole fraction.
C                 These are the UNCONSTRAINED mole fraction derivatives,
C                 the derivatives in which the mole fractions are
C                 NOT required to sum to unity.
C        dGamT  - Vector of working length NC containing the
C                 derivatives of ln phi(i) with respect to
C                 temperature.
C        dGamP  - Vector of working length NC containing the
C                 derivatives of ln phi(i) with respect to
C                 pressure
c     integer            nc, nd, Iphase
c     integer            iOpt(4)
c     integer            Kmod(6)
c     double precision   lnphi(nd)
c     double precision   dlnphi(nd,nd)
c     double precision   DPHIT(nd)
c     double precision   DPHIP(nd)
c     double precision   T, P, Z, x(nd)
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)

c     Check the availability to compute a liquid activity coefficients fo the given COPP
      iErr = IsActCof(2)
      if (iErr .ne. 0) then
c       Compute them
        call Act_Cof(LnPhiL,dLnPhi,dPhiT,dPhiP,2,iOpt,
     +               T,P,X,nap,ncmax,iErr)
        if (iErr .ne. 0) then
          write(*,*) '*** Failure in Act_Cof: error = ',iErr
        else if (io .ne. 0) then
          write(io,*) 'Act_Cof succeeded'
          do i=1,nap
            write(io,*) 'ln(gam) (',i,')=',LnPhiL(i)
          end do
          if (iOpt(2) .ne. 0) then
            do i=1,nap
              write(io,*) 'dLn(Gam)dx=',(dLnPhi(i,j),j=1,nap)
            end do
          end if
          if (iOpt(3) .ne. 0) then
            do i=1,nap
              write(io,*) 'dLn(gam)dT=',dPhiT(i)
            end do
          end if
          if (iOpt(4) .ne. 0) then
            do i=1,nap
              write(io,*) 'dLn(Gam)dP=',dPhiP(i)
            end do
          end if
        end if
      else
        write(*,*) '*** model has no liquid activity coefficients ***'
      end if

c     ------------------------------------------------------------------
c      subroutine Enthalpy(H,dHdx,dHdT,dHdP,iPhase,iOpt,Key,
c     +                    T,P,X,nc,nd,iErr)
C     Input:
C        T      - Temperature (Kelvin)
C        P      - Pressure (Pascals)
C        X      - Vector with mixture composition (mole fraction)
C        NC     - Number of components
C        iOpt   - integer vector of length 4 containing flags to
C                 indicating which properties are to be evaluated
C                 iOpt(1) > 0: Enthalpy
C                 iOpt(2) > 0: Composition derivatives
C                 iOpt(3) > 0: T derivatives
C                 iOpt(4) > 0: P derivatives
C        iPhase - Phase:
C                 1 = vapor
C                 2 = liquid
C        KEY    - integer equal to zero ONLY if K-value was called
C                 at the SAME TEMPERATURE, PRESSURE and COMPOSITION
C                 prior to the call to ENTH
C     Output:
C        H      - Enthalpy of mixture (J / kmol)
C                 Reference state is ideal gas at 298 K
C        dHdx   - Vector of working length NC containing
C                 partial derivatives of enthalpy wrt composition
C        dHdT   - Partial derivative of enthalpy wrt temperature
C        dHdP   - Partial derivative of enthalpy wrt pressure
c     integer Hmodel, Kmod(6), iOpt(4)
c     integer iPhase, I, NC, KEY
c     double precision T, P
c     double precision H, dHdT, dHdP
c     double precision dHdx(nd), x(nd)
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call Enthalpy(HL,dHdx,dHdT,dHdP,2,iOpt,1,
     +              T,P,X,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Enthalpy (L): error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Enthalpy (L) succeeded'
        write(io,*) 'T, HL=',T, HL
        if (iOpt(2) .ne. 0) then
          write(io,*) 'dHLdx=',(dHdx(j),j=1,nap)
        end if
        if (iOpt(3) .ne. 0) then
          write(io,*) 'dHLdT=',dHdT
        end if
        if (iOpt(4) .ne. 0) then
          write(io,*) 'dHLdP=',dHdP
        end if
      end if
c
      call Enthalpy(HV,dHdx,dHdT,dHdP,1,iOpt,1,
     +              T,P,Y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Enthalpy (V): error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Enthalpy (V) succeeded'
        write(io,*) 'HV=',HV
        if (iOpt(2) .ne. 0) then
          write(io,*) 'dHVdx=',(dHdx(j),j=1,nap)
        end if
        if (iOpt(3) .ne. 0) then
          write(io,*) 'dHVdT=',dHdT
        end if
        if (iOpt(4) .ne. 0) then
          write(io,*) 'dHVdP=',dHdP
        end if
      end if
c
      FFeed = Liquid
      Hfeed = HL

c     ------------------------------------------------------------------
c      subroutine Entropy(S,dSdx,dSdT,dSdP,iPhase,iOpt, T,p,x,nc,nd,iErr)
c     Input:
c        T      - Temperature (Kelvin)
c        p      - Pressure (Pascals)
c        x      - Vector with mixture composition (mole fraction)
c        nc     - Number of components
C        iPhase - Phase:
C                 1 = vapor
C                 2 = liquid
c     Output:
c        S      - Entropy of mixture (J / kmol / K)
c     integer           Smodel, Kmod(6)
c     integer           iphase, nc, nd, ierr
c     double precision  T, P, x(nd)
c     double precision  S
c     ------------------------------------------------------------------
      if (io .ne. 0) write(io,*)
      call Entropy (S,dSdx,dSdT,dSdP,2,iOpt,T,p,x,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Entropy (L): error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Entropy (L) succeeded'
        write(io,*) 'SL=',S
        if (iOpt(2) .ne. 0) then
          write(io,*) 'dSLdx=',(dSdx(j),j=1,nap)
        end if
        if (iOpt(3) .ne. 0) then
          write(io,*) 'dSLdT=',dSdT
        end if
        if (iOpt(4) .ne. 0) then
          write(io,*) 'dSLdP=',dSdP
        end if
      end if
c
      call Entropy (S,dSdx,dSdT,dSdP,1,iOpt,T,p,y,nap,ncmax,iErr)
      if (iErr .ne. 0) then
        write(*,*) '*** Failure in Entropy (V): error = ',iErr
      else if (io .ne. 0) then
        write(io,*) 'Entropy (V) succeeded'
        write(io,*) 'SV=',S
        if (iOpt(2) .ne. 0) then
          write(io,*) 'dSVdx=',(dSdx(j),j=1,nap)
        end if
        if (iOpt(3) .ne. 0) then
          write(io,*) 'dSVdT=',dSdT
        end if
        if (iOpt(4) .ne. 0) then
          write(io,*) 'dSVdP=',dSdP
        end if
      end if

c     ------------------------------------------------------------------
c     pH flash test
      iFlsh = 8
c      T=240+273.15
c      p=10.0d5
      Q = -10010.78 / Ffeed
c      Ffeed=1.d0
c      Liquid=0.d0
c      Vapour=Feed
      if (io .ne. 0) then
        write(io,*)
        write(io,*) 'ChemSep pH Flash Test'
        write(io,*) 'T =',T,' p = ',P
        write(io,*) 'F =',Ffeed,' L = ',Liquid,' V = ',Vapour
        write(io,*) 'Q =',Q
      end if
c
      call C_Flash(iFlsh,0,Q,Ffeed,
     +             iPhase,Vapour,Liquid,
     +             1,
     +             T,P,Z,x,y,nap,ncmax,iErr)

      if (iErr .ne. 0) then
        write(*,*) '*** Failure in C_Flash: error = ',iErr
        stop
      else if (io .ne. 0) then
        write(io,*) 'Calculated ChemSep pH Flash'
        write(io,*) 'T =',T,' p = ',P
        write(io,*) 'F =',Ffeed,' L = ',Liquid,' V = ',Vapour
        write(io,*) 'Q =',Q
        do i=1,nap
          write(io,*) 'i =',i,' x(i)=',x(i),' y(i)=',y(i)
        end do
        write(io,*)
      end if
c     ------------------------------------------------------------------

c     ------------------------------------------------------------------
c     Extra Properties
      nex = NumberExtraProps()
      if (io .ne. 0) write(io,*) nex,' number of extra properties'
      do j=1,nex
        i = idExtraProperty(j)
        call EPname (i, Cname)
        call EPdesc (i,iDmns,Desc)
        call EProp (i, Cvalue, T, p, Ffeed, Z, nc, ncmax, iErr)
        if (io .ne. 0) then
          if (iErr .eq. 0) then
            write(io,*) Cvalue, ' '//Cname
          else
            write(io,*) 'error calculating '//Cname
          end if
        end if
        write(io,*)
      end do
c     ------------------------------------------------------------------
c
c      call DebugOff(iErr)
c
      write(*,*) 'TESTCOPP finished successfully'
      write(*,*)
      stop
      end



      integer function StrLen (string)
      character string*(*)
      integer   i
      do i = LEN(string), 1, -1
         if (string(i:i) .ne. ' ') then
           StrLen = i
           return
         end if
      end do
      StrLen=0
      return
      end



      logical function has_substring (string, substring)
      character string*(*)
      character substring*(*)
      if (index(string,substring) .ne. 0) then
        has_substring = .true.
      else
        has_substring = .false.
      end if
      return
      end
